import typing
import pathlib
import collections

from csvw.datatypes import anyURI
from pycldf import Dataset
from pycldf.media import MediaTable, File
from pycldf.orm import Object


def as_list(obj):
    if isinstance(obj, list):
        return obj
    return [obj]


def get_objects_and_media(
        ds: Dataset,
        comp: str,
        refprop: str,
        filter: typing.Optional[typing.Callable[[Object], bool]] = None
) -> typing.List[typing.Tuple[Object, typing.List[File]]]:
    """
    Objects can be related to media via a `mediaReference` on the object's table or via an object
    reference on `MediaTable`.
    """
    objs, media = [], set()
    if (comp, 'mediaReference') in ds:
        for obj in ds.objects(comp):
            if filter is None or filter(obj):
                mrefs = as_list(obj.cldf.mediaReference)
                media |= set(mrefs)
                objs.append((obj, mrefs))
    elif ('MediaTable', refprop) in ds:
        media_by_fid = collections.defaultdict(list)
        for row in ds.iter_rows('MediaTable', 'id', refprop):
            for fid in as_list(row[refprop]):
                media_by_fid[fid].append(row['id'])
        for obj in ds.objects(comp):
            if filter is None or filter(obj):
                mrefs = media_by_fid.get(obj.id, [])
                media |= set(mrefs)
                objs.append((obj, mrefs))
    else:
        objs = [(obj, []) for obj in ds.objects(comp) if filter is None or filter(obj)]

    media = {mid: None for mid in media}
    if 'MediaTable':
        for file in MediaTable(ds):
            if file.id in media:
                media[file.id] = file
    return [(obj, [media[mid] for mid in mrefs if media[mid]]) for obj, mrefs in objs]


def get_media_url(
        file: File, media_dir: typing.Optional[pathlib.Path] = None) -> typing.Union[None, str]:
    if media_dir:
        if file.local_path(media_dir).exists():
            # Read audio from the file system:
            return 'file://{}'.format(file.local_path(media_dir).resolve())
    else:
        # Read audio from the URL:
        return anyURI.to_string(file.url)



"""
This module provides shared functionality for cli commands.

Often, functionality comes as two related functions `add_*` and `get_*`, where the `add_*`
function registers options which are exploited in the `get_*` function.
"""
import re
import json
import typing
import inspect
import pathlib
import argparse
import webbrowser

from clldutils.text import split_text_with_context
from clldutils.clilib import PathType, ParserError
from clldutils import path
import newick
from pyglottolog.objects import Glottocode
from pycldf import Dataset
from pycldf.ext import discovery
from pycldf.trees import TreeTable, Tree

from cldfviz.glottolog import Glottolog
from cldfviz.colormap import COLORMAPS, CATEGORICAL, CONTINUOUS, Colormap
from cldfviz.multiparameter import MultiParameter


def join_quoted(items: typing.Iterable) -> str:
    return ', '.join(['"{}"'.format(i) for i in items])


def add_jinja_template(parser, default):
    parser.add_argument(
        "--template",
        type=PathType(type='file'),
        default=default,
        help="Template file using Jinja2 syntax (see https://jinja.palletsprojects.com/). "
             "To create a custom template, you might want to start out with a copy of the "
             "default template.",
    )


def add_open(parser):
    parser.add_argument(
        "--open", action='store_true', default=False,
        help="Open the output file in the browser. (Requires specifying --output as well.)"
    )
    try:
        parser.add_argument(
            "-o", "--output", type=PathType(type="file", must_exist=False), default=False,
            help="(non-existing) path name to write the result to.",
        )
    except argparse.ArgumentError:  # pragma: no cover
        pass  # output option already added.


def open_output(args: argparse.Namespace):
    if args.output and args.open and not getattr(args, 'test', False):  # pragma: no cover
        webbrowser.open(args.output.resolve().as_uri(), new=1)


def write_output(args: argparse.Namespace, res: str):
    if args.output:
        args.output.write_text(res, encoding="utf8")
        print("Output written to {}".format(args.output))
        open_output(args)
    else:
        print(res)


def add_language_filter(parser):
    parser.add_argument(
        '--language-filters',
        default=None,
        help="JSON object specifying filter criteria for included languages. Keys must be "
             "names of columns in the datasets' LanguageTable, values are interpreted as "
             "regular expressions if they are strings or as literal values otherwise and will "
             "be matched against the *string* value of a language for the specified column "
             '(thus, to match boolean values, the filter must specify "False" or "True"). Only '
             "languages matching all criteria are included in the analysis.",
    )


def get_language_filter(args):
    if args.language_filters is None:
        return

    def language_filter(lg):
        for k, v in json.loads(args.language_filters).items():
            val = lg.data[k]
            if isinstance(v, str):
                if isinstance(val, list):
                    if v not in val:
                        return False
                elif not re.search(v, str(val) or ''):
                    return False
            else:
                if lg.data[k] != v:
                    return False
        return True
    return language_filter


def get_filtered_languages(args, ds) -> typing.Union[None, typing.List[str]]:
    language_filter = get_language_filter(args)
    if language_filter:
        res = []
        if 'LanguageTable' not in ds:  # pragma: no cover
            raise ValueError('Language filters only work on datasets with a LanguageTable')
        for lg in ds.objects('LanguageTable'):
            if language_filter(lg):
                res.append(lg.id)
        return res


def add_testable(parser):
    parser.add_argument('--test', action='store_true', default=False, help=argparse.SUPPRESS)


def add_listvalued(parser, *args, **kw):
    kw.setdefault('default', [])
    kw.setdefault('type', lambda s: split_text_with_context(s, ',', brackets={'{': '}'}))
    parser.add_argument(*args, **kw)


def import_module(dotted_name_or_path):
    import importlib
    p = pathlib.Path(dotted_name_or_path)
    if p.exists():
        return path.import_module(p.resolve())
    return importlib.import_module(dotted_name_or_path)


def import_subclass(dotted_name_or_path, cls):
    mod = import_module(dotted_name_or_path)
    for _, obj in inspect.getmembers(mod):
        if inspect.isclass(obj) and issubclass(obj, cls) and not obj.__subclasses__():
            return obj


def add_multiparameter(parser, with_language_filter=False, with_language_properties=False):
    """
    Adds necessary options to instantiate a `MultiParameter` object (with related colormaps).
    """
    if with_language_filter:
        add_language_filter(parser)
    add_listvalued(
        parser,
        '--parameters',
        help="Comma-separated Parameter IDs, specifying the values to plot.",
    )
    add_listvalued(
        parser,
        '--datatypes',
        help="Explicit datatypes for parameters",
    )
    add_listvalued(
        parser,
        '--colormaps',
        help="Comma-separated names of colormaps to use for the respective parameter. Choose from "
             "{} for categorical and from {} for continuous parameters."
             "".format(join_quoted(COLORMAPS[CATEGORICAL]), join_quoted(COLORMAPS[CONTINUOUS])),
    )
    parser.add_argument(
        '--missing-value',
        default=None,
        help="A color used to indicate missing values. If not specified missing values will be "
             "omitted. (Note: We define 'missing values' as rows in the ValueTable with a null "
             "value; so specifying --missing-value will **not** add null values for all languages "
             "in the dataset.)",
    )
    parser.add_argument(
        '--weight-col',
        help="Name of a column in ValueTable with numeric values to use as weight (for "
             "multi-valued parameters).",
        default=None,
    )
    if with_language_properties:
        add_listvalued(
            parser,
            '--language-properties',
            help="Comma-separated language properties, i.e. columns in the dataset's LanguageTable "
                 "to plot on the map.",
        )
        add_listvalued(
            parser,
            '--language-properties-colormaps',
            help="Comma-separated names of colormap to use for the respective language properties. "
                 "See help for --colormaps for choices",
        )


def get_multiparameter(args, ds: Dataset, glottolog: Glottolog, **ukw):
    with_language_filters = hasattr(args, 'language_filters')
    with_language_properties = hasattr(args, 'language_properties')

    kw = dict(
        datatypes=args.datatypes,
        glottolog=glottolog,
        include_missing=args.missing_value is not None,
        weight_col=getattr(args, 'weight_col', None),
    )
    if with_language_properties:
        kw['language_properties'] = args.language_properties
    if with_language_filters:
        kw['language_filter'] = get_language_filter(args)
    kw.update(ukw)

    data = MultiParameter(ds, args.parameters, **kw)

    if args.parameters and not args.colormaps:
        args.colormaps = [None] * len(args.parameters)

    if with_language_properties:
        if args.language_properties and not args.language_properties_colormaps:
            args.language_properties_colormaps = \
                [None] * len(args.language_properties)
        if '__language__' in data.parameters:
            assert len(data.parameters) == 1
            args.colormaps = args.colormaps or [None]
        args.colormaps.extend(args.language_properties_colormaps)

    assert len(args.colormaps) == len(data.parameters), '{}'.format(data.parameters.keys())
    try:
        cms = {
            pid: Colormap(
                data.parameters[pid],
                name=cm,
                novalue=args.missing_value)
            for pid, cm in zip(data.parameters, args.colormaps)}
    except (ValueError, KeyError) as e:
        raise ParserError(str(e))

    with_shapes = sum(1 for cm in cms.values() if cm.with_shapes)
    if with_shapes:
        if with_shapes > 1:
            raise ParserError('Only one colormap can specify shapes.')
        if len(data.parameters) > 2:
            raise ParserError('Shapes can only be specified for one of two parameters.')

    return data, cms


def add_secondary_dataset(parser, opt: str, help: typing.Optional[str] = None):
    """
    Some commands access data in multiple CLDF datasets.

    To be used with `get_secondary_dataset`.
    """
    parser.add_argument(
        opt,
        default=None,
        help="CLDF dataset locator. {}".format(help or '').strip(),
    )
    try:
        parser.add_argument(
            '--download-dir',
            type=PathType(type='dir'),
            help='An existing directory to use for downloading a dataset (if necessary).',
            default=None,
        )
    except argparse.ArgumentError:  # pragma: no cover
        pass  # output option already added.


def add_tree(parser):
    add_secondary_dataset(
        parser,
        '--tree-dataset',
        help='A tree from the TreeTable of this dataset is used.')
    parser.add_argument(
        '--tree',
        help="Tree specified as Glottocode (interpreted as the root of the Glottolog tree), "
             "Newick formatted string or path to a file containing the Newick formatted "
             "tree.")
    parser.add_argument(
        '--tree-id', default=None,
        help="Tree specified by ID in the TreeTable of a CLDF dataset."
    )
    parser.add_argument(
        '--glottocodes-as-tree-labels',
        action='store_true',
        default=False,
        help="If a tree in a TreeTable of a CLDF dataset is used, the nodes will be renamed "
             "using the corresponding Glottocodes."
    )


def get_secondary_dataset(args, opt: str):
    if getattr(args, opt):
        return discovery.get_dataset(getattr(args, opt), args.download_dir)


def get_tree(args, glottolog: typing.Optional[Glottolog] = None) \
        -> typing.Tuple[newick.Node, typing.Union[None, Tree], typing.Union[None, Dataset]]:
    """
    Note: While the `Tree` object which may be returned by this function allows access to the
    Newick string of the tree, the returned `Node` object should be used preferentially. This is
    because changes like renaming labels to Glottocodes are only done on this `Node` and **not**
    synced to `Tree.newick_string`!
    """
    ds = get_secondary_dataset(args, 'tree_dataset')
    tree = None
    if args.tree:
        if Glottocode.pattern.match(args.tree):
            assert glottolog
            nwk = glottolog.newick(args.tree)
        elif pathlib.Path(args.tree).exists():
            nwk = newick.read(args.tree)[0]
        else:
            nwk = newick.loads(args.tree)[0]
    else:
        for tree in TreeTable(ds):
            if (args.tree_id and tree.id == args.tree_id) or \
                    ((not args.tree_id) and tree.tree_type == 'summary'):
                nwk = tree.newick()
                break
        else:
            raise ValueError('No matching tree found')  # pragma: no cover
        # Rename tree nodes:
        if args.glottocodes_as_tree_labels:
            name_map = {
                r['id']: r['glottocode']
                for r in ds.iter_rows('LanguageTable', 'id', 'glottocode')}

            def rename(n):
                n.name = name_map.get(n.name)
                return n
            nwk.visit(rename)
    return nwk, tree, ds



"""
Plots a phylogeny as SVG.
"""
import types
import pathlib

from cldfviz.cli_util import (
    add_testable, add_language_filter, get_language_filter, add_open, write_output,
    get_multiparameter, add_multiparameter, add_tree, get_tree,
    add_secondary_dataset, get_secondary_dataset,
)
from cldfviz.glottolog import Glottolog
from cldfviz.colormap import weighted_colors
from cldfviz.tree import render


def register(parser):
    add_testable(parser)
    add_tree(parser)
    add_language_filter(parser)
    Glottolog.add(parser)
    parser.add_argument(
        '--glottolog-links',
        help="Turn language labels into links to Glottolog (where possible).",
        action='store_true',
        default=False)
    parser.add_argument(
        '--ascii-art',
        help="Only print tree as ASCII graphic to the terminal.",
        action='store_true',
        default=False)
    parser.add_argument(
        '--name-as-label',
        action='store_true',
        default=False)
    parser.add_argument('--title', default=None)
    parser.add_argument('--width', type=int, default=500)
    parser.add_argument('--height', type=int, default=None)
    parser.add_argument(
        '--styles',
        help="Python dict suitable to pass into `toytree.tree.draw` (or path to a text file "
             "containing such a dict). "
             "See https://toytree.readthedocs.io/en/latest/8-styling.html#",
        default='{}')
    add_secondary_dataset(parser, '--data-dataset')
    parser.add_argument(
        '--tree-label-property',
        help="Name of the language property used to identify languages in the tree.",
        default=None)
    add_multiparameter(parser)
    add_open(parser)


def run(args):
    cldf = get_secondary_dataset(args, 'data_dataset')
    nwk, tree, treeds = get_tree(args, glottolog=Glottolog.from_args(args))

    if args.ascii_art:
        print(nwk.ascii_art())
        return

    data = None
    if args.parameters:
        mp, cms = get_multiparameter(args, cldf, None)
        values = {lang.id: weighted_colors(val, cms) for lang, val in mp.iter_languages()}
        data = types.SimpleNamespace(values=values, parameters=mp.parameters, colormaps=cms)

    if args.title:
        legend = args.title
    else:
        legend = tree.name if tree else ''
        if treeds:
            dcol = treeds.get(('TreeTable', 'description'))
            if dcol:
                legend += '{}{}'.format(' - ' if legend else '', tree.row[dcol.name])
        if tree and tree.tree_branch_length_unit:
            legend += ' with branches in {}'.format(tree.tree_branch_length_unit)

    glangs = {}
    if args.glottolog and args.glottolog_links:  # pragma: no cover
        glangs = {lg.id: lg.name for lg in args.glottolog.api.languoids()}
    if pathlib.Path(args.styles).exists():
        args.styles = pathlib.Path(args.styles).read_text(encoding='utf8')
    lf = get_language_filter(args)

    lid2tree = {}
    lid2name = {}
    if cldf:
        # We need to collect two mappings here:
        # 1. Language ID to node labels used in the tree.
        # 2. Language ID to names, in case we need to rename the leaf nodes.
        for lg in cldf.iter_rows('LanguageTable', 'id', 'glottocode', 'name'):
            lid2tree[lg['id']] = \
                lg[args.tree_label_property] if args.tree_label_property else lg['id']
            lid2name[lg['id']] = lg['name']

    if data and args.tree_label_property:
        # Re-key the values so that they can be picked up when adding markers to the tree.
        data.values = {lid2tree[lid]: vals for lid, vals in data.values.items() if lid2tree[lid]}

    labels = {}
    if args.name_as_label:
        if cldf:
            # If we have data, we assume that the tree nodes should get data language names.
            for lid, tid in lid2tree.items():
                labels[tid] = lid2name[lid]
        elif treeds:
            # If we have a tree from a tree dataset, "name-as-label" refers to language names in
            # this dataset.
            labels = {
                r['id']: r['name']
                for r in treeds.iter_rows('LanguageTable', 'id', 'name')
            }

    kw = dict(
        legend=legend,
        width=args.width,
        height=args.height,
        styles=eval(args.styles),
        with_glottolog_links=args.glottolog_links,
        data=data,
        labels=labels or None,
    )
    if treeds:
        kw.update(
            tree_object=tree,
            glottolog_mapping={
                r['id']: (r['glottocode'], glangs.get(r['glottocode']) or '') for r in
                treeds.iter_rows('LanguageTable', 'id', 'glottocode') if r['glottocode']},
            leafs=[lg.id for lg in treeds.objects('LanguageTable') if lf(lg)] if lf else None,
        )
    write_output(args, render(nwk, **kw))



"""
Convert a parameter network to a suitable format for visualization/analysis.

- DOT/GV for interoperability with the GraphViz tools
- GraphML for interoperability with networkx/igraph/Gephi

Image formats like SVG can then be created by piping the output to the `dot` program.
"""
import re
import json
import pathlib
import functools
import xml.etree
import collections

try:
    from networkx import Graph
    from networkx.algorithms.components import node_connected_component
except ImportError:  # pragma: no cover
    Graph = None

from pycldf.cli_util import get_dataset, add_dataset


def string_or_path(s):
    return pathlib.Path(s).read_text(encoding='utf8') if pathlib.Path(s).exists() else s


def register(parser):
    add_dataset(parser)
    parser.add_argument(
        '--format',
        choices=['dot', 'graphml'],
        help="Output format",
        default='dot')
    parser.add_argument(
        '--edge-filters',
        default=None,
        help="JSON object specifying filter criteria for included edges. Keys must be "
             "names of columns in the datasets' ParameterNetwork, values are interpreted as "
             "regular expressions if they are strings or as literal values otherwise and will "
             "be matched against the value of an edge for the specified column. Only "
             "edges matching all criteria are included in the analysis.",
    )
    parser.add_argument(
        '--node-attributes',
        help="Python dict (passed as string or name to a file) to supply node attributes. Use "
             "lambda functions accepting a pycldf.orm.Parameter instance as sole argument to "
             "compute values on the fly.",
        type=string_or_path,
        default='{}')
    parser.add_argument(
        '--edge-attributes',
        help="Python dict (passed as string or name to a file) to supply edge attributes. Use "
             "lambda functions accepting a pycldf.orm.ParameterNetworkEdge instance as sole "
             "argument to compute values on the fly.",
        type=string_or_path,
        default='{}')
    parser.add_argument(
        '--parameter',
        help="To single out individual components of multi-component networks, a node contained "
             "in the desired component can be specified by ID or Name.",
        default=None,
    )


def run(args):
    if Graph is None:  # pragma: no cover
        args.log.error(
            'install cldfviz with network support, running "pip install cldfviz[network]"')
        return

    def edge_filter(row):
        for k, v in json.loads(args.edge_filters or '{}').items():
            val = row.data[k]
            if isinstance(v, list):
                if val not in v:
                    return False
            elif isinstance(v, str):
                if not re.search(v, str(val) or ''):
                    return False
            else:
                if row.data[k] != v:
                    return False
        return True

    ds = get_dataset(args)
    pnodes = collections.OrderedDict((p.id, p) for p in ds.objects('ParameterTable'))

    # Edges filtered by --edge-filters:
    edges = [edge for edge in ds.objects('ParameterNetwork') if edge_filter(edge)]

    if args.parameter:
        # Now compute edges filtered for one component:
        g = Graph()
        for e in edges:
            g.add_edge(e.cldf.sourceParameterReference, e.cldf.targetParameterReference)
        for pid, p in pnodes.items():
            if args.parameter == p.cldf.name or args.parameter == pid:
                nodes = node_connected_component(g, pid)
                edges = [e for e in edges if e.cldf.sourceParameterReference in nodes]
                break
        else:  # pragma: no cover
            raise ValueError('Invalid parameter specified.')

    nids = ({e.cldf.sourceParameterReference for e in edges} |  # noqa: W504
            {e.cldf.targetParameterReference for e in edges})
    nodes = [v for k, v in pnodes.items() if k in nids]

    def compute_attrs(spec, obj):
        attrs, drop = collections.OrderedDict(), False
        for k, v in spec.items():
            if callable(v):
                v = v(obj)
                if k == 'drop' and v:
                    # Special key "drop" and the return value is True, skip the edge.
                    drop = True
                    break
            if k != 'drop':
                attrs[k] = v
        return attrs, drop

    # Compute attributes:
    eattrs = functools.partial(compute_attrs, eval(args.edge_attributes))
    edges = [(edge, eattrs(edge)) for edge in edges]
    edges = [(edge, attrs) for edge, (attrs, drop) in edges if not drop]
    nattrs = functools.partial(compute_attrs, eval(args.node_attributes))
    nodes = [(node, nattrs(node)[0]) for node in nodes]

    if args.format == 'dot':
        print(dot(nodes, edges))
    elif args.format == 'graphml':
        print(graphml(nodes, edges))
    else:  # pragma: no cover
        raise ValueError('unsupported format')


def dot(nodes, edges):
    def dot_id(v):
        """
        We convert everything to a string, then escape as needed.
        """
        return '"{}"'.format(str(v).replace('"', r'\"'))

    def format_attrs(attrs):
        pairs = ['"{}"={}'.format(k, dot_id(v)) for k, v in attrs.items() if v is not None]
        return ' [{}]'.format(';'.join(pairs)) if pairs else ''

    def iter_lines():
        yield 'digraph {'
        for node, attrs in nodes:
            attrs.setdefault('label', node.cldf.name)
            yield '"{}"{}'.format(node.id, format_attrs(attrs))

        for edge, attrs in edges:
            if not edge.cldf.edgeIsDirected:
                attrs.setdefault('dir', 'none')
            yield '"{}" -> "{}"{}'.format(
                edge.cldf.sourceParameterReference,
                edge.cldf.targetParameterReference,
                format_attrs(attrs))
        yield '}'
    return '\n'.join(iter_lines())


def graphml(nodes, edges, edgedefault="undirected"):
    ns = 'http://graphml.graphdrawing.org/xmlns'

    def qname(lname):
        return '{' + ns + '}' + lname

    def infer_type(values):
        #  boolean, int, long, float, double, or string.
        for value in values:
            if isinstance(value, bool):
                return 'boolean', lambda s: str(s).lower()
            if isinstance(value, int):
                return 'long', lambda s: str(s)
            if isinstance(value, float):
                return 'double', lambda s: str(s)
        return 'string', lambda s: str(s)

    def iter_attributes(attrs):
        if attrs:
            values = collections.defaultdict(list)
            for attr in attrs:
                for k, v in attr.items():
                    values[k].append(v)
            for key, vals in values.items():
                yield key, infer_type(vals)

    def subelement(parent, tag, **attrs):
        return xml.etree.ElementTree.SubElement(
            parent, qname(tag), **{qname(k): v for k, v in attrs.items()})

    def create_key(e, prefix, name, type):
        id_ = '{}.{}'.format(prefix, name)
        subelement(e, 'key', **{'id': id_, 'for': 'node', 'attr.name': name, 'attr.type': type})
        return id_

    root = xml.etree.ElementTree.Element(qname('graphml'))
    graph = subelement(root, 'graph', id='graph', edgedefault=edgedefault)
    converters = {}
    for aname, (atype, conv) in iter_attributes([a for _, a in nodes]):
        converters[create_key(graph, 'n', aname, atype)] = conv
    for aname, (atype, conv) in iter_attributes([a for _, a in edges]):
        converters[create_key(graph, 'e', aname, atype)] = conv

    for node, attrs in nodes:
        n = subelement(graph, 'node', id=node.id)
        for k, v in attrs.items():
            key = 'n.{}'.format(k)
            if v is not None:
                t = subelement(n, 'data', key=key)
                t.text = converters[key](v)

    for edge, attrs in edges:
        n = subelement(graph, 'edge',
                       source=edge.cldf.sourceParameterReference,
                       target=edge.cldf.targetParameterReference)
        for k, v in attrs.items():
            key = 'e.{}'.format(k)
            if v is not None:
                t = subelement(n, 'data', key=key)
                t.text = converters[key](v)

    if hasattr(xml.etree.ElementTree, 'indent'):  # new in Python 3.9.
        xml.etree.ElementTree.indent(root)
    return xml.etree.ElementTree.tostring(
        root,
        encoding='utf8',
        default_namespace='http://graphml.graphdrawing.org/xmlns',
        xml_declaration=True).decode('utf8')



import itertools

import pytest

from cldfviz.multiparameter import Parameter, Value
from cldfviz.colormap import *


@pytest.mark.parametrize(
    's,res',
    [
        ('r', '#FF0000'),
        ('aqua', '#00FFFF'),
        ('a00', '#AA0000'),
        ('diamond', 'diamond'),
    ]
)
def test_hextriplet(s, res):
    assert hextriplet(s) == res


@pytest.mark.parametrize(
    'vals,kw,val,expected',
    [
        ((1, 5), {}, 2, '#0080ff'),
        ([1, '4', 5], {'name': 'tol'}, 5, '#CC6677'),
    ]
)
def test_Colormap(vals, kw, val, expected):
    cm = Colormap(Parameter(id='x', name='y', domain=vals), **kw)
    assert cm(val) == expected
    assert cm.scalar_mappable()


def test_Colormap2():
    with pytest.raises(ValueError):
        _ = Colormap(Parameter(id='x', name='y', domain=dict(a=1)), name='seq', novalue='x')

    param = Parameter(
        id='x',
        name='y',
        domain=dict(a=1, b=2, c=3),
        value_to_code={'a': 'a', 'b': 'b', 'c': 'c', 'None': None},
    )
    with pytest.raises(ValueError):
        _ = Colormap(param, name='{"a":"0a0"}', novalue='a00')

    cm = Colormap(param, name='{"a":"0a0","b":"fec44f","c":"blue"}', novalue='a00')
    assert cm('b') == '#FEC44F'
    assert cm(None) == '#AA0000'
    cm = Colormap(param, name='seq')
    assert cm('b') == '#FEC44F'


def test_get_shape_and_color():
    with pytest.raises(ValueError):
        get_shape_and_color([(1, 'circle'), (1, 'diamond')])
    assert get_shape_and_color([(1, '#aaaaaa'), (1, 'circle')]) == ('circle', '#aaaaaa')


def _make_value(v, pid, weight=None):
    return Value(v=v, pid=pid, lid='l', code=v, weight=weight)


@pytest.mark.parametrize(
    'values,count',
    [
        ([_make_value('x', 'a')], 1),
        ([_make_value('x', 'a'), _make_value('x', 'a')], 1),
        ([_make_value('x', 'a'), _make_value('y', 'a')], 2),
        ([_make_value('x', 'a'), _make_value('y', 'b'), _make_value('x', 'b', 3)], 3),
        ([_make_value('x', 'a', 20), _make_value('y', 'a', 2)], 2),
        ([_make_value('x', 'a'), _make_value('x', 'b')], 2),
    ]
)
def test_weighted_colors(values, count):
    class Colormap(dict):
        def __call__(self, v):
            return self[v]

    values = {
        pid: list(vals) for pid, vals in
        itertools.groupby(sorted(values, key=lambda v: v.pid), lambda v: v.pid)}
    colormaps = dict(a=Colormap(x='x', y='y'), b=Colormap(x='x', y='y', z='z'))
    res = weighted_colors(values, colormaps)
    assert len(res) == count
    assert sum(c[0] for c in res) == pytest.approx(1.0)



"""
Plot values for a parameter of a CLDF dataset on a map linked to a tree.

The tree can be specified in various ways:

1. Using the `--tree` option which accepts
   - a Glottocode - in which the Glottolog sub-classification tree for the languoid specified by
     `--tree` is used,
   - a Newick-formatted string,
   - a path name to an existing file containing the Newick-formatted tree.
2. Using the `--tree-dataset` (and `--tree-id`) options to select a tree provided in the `TreeTable`
   of a CLDF dataset.

Linking the languages in the parameter's dataset to languages in the tree can be configured via the
`--glottocodes-as-tree-labels` and `--tree-label-property` options. The former will re-name tree
labels to Glottocodes using the mapping given in the `--tree-dataset`'s `LanguageTable`. The latter
specifies a column in the parameter's `LanguageTable` values of which are used to match languages to
tree node labels.

The actual map rendering is done by `lingtreemaps.plot` and all options prefixed with `--ltm-` are
passed into this function as keyword arguments. Thus, for map customization consult the docs at
https://lingtreemaps.readthedocs.io/en/latest/config.html
"""
import io
import logging
import pathlib
import collections

from pycldf.cli_util import add_dataset, get_dataset

from cldfviz.cli_util import (
    add_testable, add_open, open_output, add_language_filter, get_filtered_languages,
    add_tree, get_tree,
)
from cldfviz.glottolog import Glottolog
from cldfviz.pdutils import df_from_dicts

try:
    from Bio import Phylo
    import yaml
    import lingtreemaps
except ImportError:  # pragma: no cover
    lingtreemaps = None


def yaml_type(s):
    return yaml.load(io.StringIO(s), yaml.SafeLoader)


def iter_ltm_options():
    h = []
    if lingtreemaps:
        for line in pathlib.Path(lingtreemaps.__file__).parent\
                .joinpath('data', 'default_config.yaml').read_text(encoding='utf8').split('\n'):
            line = line.strip()
            if line.startswith('#'):
                h.append(line[1:].strip())
            elif line:
                opt, _, default = line.partition(':')
                opt, val, help = opt.strip(), yaml_type(default.strip()), ' '.join(h)
                if opt == 'filename':
                    help = "The filename. If unspecified, the parameter ID will be used."
                yield opt, val, help
                h = []


def register(parser):
    add_testable(parser)
    add_dataset(parser)
    Glottolog.add(parser)
    parser.add_argument('parameter')

    add_tree(parser)

    parser.add_argument(
        '--tree-label-property',
        help="Name of the language property used to identify languages in the tree.",
        default='glottocode')
    add_language_filter(parser)
    for opt, default, help in iter_ltm_options():
        parser.add_argument(
            '--ltm-{}'.format(opt.replace('_', '-')),
            default=default,
            type=yaml_type,
            help=help)
    add_open(parser)


def run(args):
    if lingtreemaps is None:  # pragma: no cover
        args.log.error(
            'install cldfviz with lingtreemaps, running "pip install cldfviz[lingtreemaps]"')
        return

    assert args.tree or args.tree_dataset
    glottolog = Glottolog.from_args(args)

    ds = get_dataset(args)
    filtered_languages = get_filtered_languages(args, ds)

    # 1. Get all values for the selected parameter:
    cols = ['parameterReference', 'languageReference', 'value']
    if ds.get(('ValueTable', 'codeReference')):
        cols.append('codeReference')
    values = {
        v['languageReference']: v for v in ds.iter_rows('ValueTable', *cols)
        if v['parameterReference'] == args.parameter and  # noqa: W504
        v['value'] and  # noqa: W504  We don't handle missing data
        (filtered_languages is None or (v['languageReference'] in filtered_languages))}

    # 2. Get the tree ...
    tree, _, _ = get_tree(args, glottolog=glottolog)
    # ... and its set of node labels.
    nodes = {n.name for n in tree.walk()}

    # 3. Get the set of matching languages in the dataset.
    treelabel2id = {}  # Maps tree label to language ID for languages in both, dataset and tree.
    languages = {}  # A dict of language-data dicts
    if 'LanguageTable' in ds:
        for lang in ds.iter_rows(
                'LanguageTable', 'id', 'glottocode', 'name', 'latitude', 'longitude'):
            if lang['id'] in values:
                # Determine which tree node name corresponds to the language:
                treelabel = lang[args.tree_label_property]
                if treelabel and treelabel in nodes:  # The language is in the tree.
                    languages[lang['id']] = lang
                    treelabel2id[treelabel] = lang['id']
    else:
        assert glottolog
        for lid in values:
            if lid in nodes:
                glang = glottolog[lid]
                languages[lid] = dict(name=lid, latitude=glang.lat, longitude=glang.lon)
                treelabel2id[lid] = lid

    for lid in list(values.keys()):
        if lid not in languages:
            del values[lid]

    # Now all values correspond to languages which are in the tree!
    assert set(values) == set(languages)

    # 4. Now we prune the tree to contain just leafs for which we have data.
    keep = {n.name for n in tree.walk() if n.name in treelabel2id}
    tree.prune_by_names(list(keep), inverse=True)
    if sum(1 for _ in tree.walk()) == 1:  # pragma: no cover
        raise ValueError('No overlap between dataset and tree')

    # 5. Rename the tree nodes
    def rename(n):
        if n.is_leaf:
            # FIXME: only quote if necessary!?
            n.name = "'{}'".format(languages[treelabel2id[n.name]]['name'])
        else:
            n.name = None
        return n
    tree.visit(rename)

    leafs = {n.unquoted_name for n in tree.walk() if n.is_leaf}
    assert leafs.issubset(set(languages[v]['name'] for v in values))
    # lingtreemaps cannot handle the case where we have values for both a language and a dialect
    # of this language. Thus, we only keep data for actual leafs.
    nonleafs = []
    for lid, lang in languages.items():
        if lang['name'] not in leafs:
            nonleafs.append(lid)
    for lid in nonleafs:
        del languages[lid]
        del values[lid]

    # 6. Turn languages and values into the appropriate data structures for lingtreemaps:
    if 'CodeTable' in ds:
        codes = collections.OrderedDict([
            (c['id'], c['name'])
            for c in ds.iter_rows('CodeTable', 'id', 'name', 'parameterReference')
            if c['parameterReference'] == args.parameter])
    else:  # pragma: no cover
        codes = None

    # Sort values by value/code. ltm will just use this order.
    values = sorted(
        values.values(), key=lambda v: v['value'] if not codes else codes[v['codeReference']])
    values = df_from_dicts(
        dict(
            Clade=languages[v['languageReference']]['name'],
            Value=v['value'] if not codes else codes[v['codeReference']]) for v in values)
    languages = df_from_dicts(
        dict(ID=lg['name'], Latitude=lg['latitude'], Longitude=lg['longitude'])
        for lg in languages.values())

    kwargs = {k.replace('ltm_', ''): v for k, v in args.__dict__.items() if k.startswith('ltm_')}
    kwargs['filename'] = kwargs['filename'] or str(args.output) or args.parameter
    args.output = pathlib.Path(
        kwargs['filename'] if '.' in kwargs['filename']
        else '{filename}.{file_format}'.format(**kwargs))

    logging.getLogger('matplotlib.font_manager').setLevel(logging.ERROR)
    lingtreemaps.plot(languages, Phylo.read(io.StringIO(tree.newick), 'newick'), values, **kwargs)
    open_output(args)



import shlex
import logging
import pathlib
import warnings
import functools

import pytest
import requests_mock
import pycldf

from cldfbench.__main__ import main

from cldfviz.map import WITH_CARTOPY, MarkerFactory, leaflet, mpl


def lmain(*args, **kw):
    kw.setdefault('log', logging.getLogger(__name__))
    return main(*args, **kw)


def runcli(command, cli):
    return lmain([command] + shlex.split(cli))


@pytest.fixture
def ds_arg(StructureDataset):
    return str(StructureDataset.directory / 'StructureDataset-metadata.json')


def test_audiowordlist(Wordlist, capsys, tmp_path):
    def run(opts):
        with warnings.catch_warnings():
            warnings.filterwarnings(
                'ignore', category=DeprecationWarning, module='importlib._bootstrap')
            warnings.filterwarnings(
                'ignore', category=DeprecationWarning, module='joblib.backports')
            functools.partial(runcli, 'cldfviz.audiowordlist')(opts)
        out, _ = capsys.readouterr()
        return out

    tmp_path.joinpath('1.wav').write_text('x', encoding='utf8')
    assert '1.wav' in run('{} ID=island --media-dir {}'.format(Wordlist.directory, tmp_path))

    assert 'example.org' in run('{} ID=island'.format(Wordlist.directory))

    run('{} ID=island --output {}'.format(Wordlist.directory, tmp_path / 'o.html'))
    assert tmp_path.joinpath('o.html').exists()

    ds = pycldf.Wordlist.in_dir(tmp_path)
    ds.add_columns(
        'FormTable',
        {'name': 'Media_ID', 'propertyUrl': 'http://cldf.clld.org/v1.0/terms.rdf#mediaReference'})
    ds.add_component('MediaTable')
    ds.write(
        FormTable=[dict(ID='1', Language_ID='l1', Parameter_ID='p1', Form='abc', Media_ID='m1')],
        MediaTable=[dict(ID='m1', Media_Type='audio/x-wav', Download_URL='http://example.org')],
    )
    assert 'abc' in run('{} p1'.format(tmp_path / 'Wordlist-metadata.json'))


def test_tree_misc(ds_arg, tmp_path, capsys):
    with warnings.catch_warnings():
        warnings.filterwarnings(
            'ignore', category=DeprecationWarning, module='importlib._bootstrap')

        styles = tmp_path / 's.json'
        styles.write_text('{}', encoding='utf8')
        o = tmp_path / 'test2.svg'
        lmain(['cldfviz.tree', '--tree-dataset', ds_arg, '--output', str(o), '--test', '--styles', str(styles)])
        assert o.exists()


@pytest.mark.parametrize(
    'args,expect',
    [
        (
            '--tree-dataset DATASET --ascii-art --tree-id 1',
            lambda out: 'Marathi' in out),
        (
            '--tree-dataset DATASET --title "The Title"',
            lambda out: 'The Title' in out),
        (
            '--tree-dataset DATASET --name-as-label',
            lambda out: 'Korku_NM' not in out and ('Korku' in out)),
        (
            '--tree "((khr:1,sat:1.1),(unr:2,bfw:1.9)):3" --tree-label-property ISO639P3code '
            '--name-as-label --data-dataset DATASET --parameters C',
            lambda out: 'Kharia' in out),
        (
            '--tree-dataset DATASET --name-as-label --data-dataset DATASET --parameters C',
            lambda out: 'Enclitic PL' in out and ('Korku_NM' not in out) and ('Korku' in out)),
        (
            '--tree-dataset DATASET --data-dataset DATASET --parameters C '
            '--colormaps \'{"0":"circle","1":"triangle_up","2":"triangle_down"}\'',
            lambda out: 'Enclitic PL' in out),
        (
            '--tree-dataset DATASET --data-dataset DATASET --parameters C,D '
            '--colormaps \'{"0":"circle","1":"triangle_up","2":"triangle_down"},tol\'',
            lambda out: 'Enclitic PL' in out),
        (
            '--tree "((Santali_NM:1,Mundari_NM:1.1),(Hindi_IA:2,Sadri_IA:1.9)):3" '
            '--data-dataset DATASET --parameters C,B',
            lambda out: 'Hindi_IA' in out),
    ]
)
def test_tree(ds_arg, tmp_path, capsys, args, expect):
    with warnings.catch_warnings():
        warnings.filterwarnings(
            'ignore', category=DeprecationWarning, module='importlib._bootstrap')

        runcli('cldfviz.tree', '--test ' + args.replace('DATASET', ds_arg))
        out, _ = capsys.readouterr()
        assert expect(out)


@pytest.mark.skipif(not WITH_CARTOPY, reason="Cannot run without cartopy.")
@pytest.mark.parametrize(
    'with_full_dataset,opts,expect',
    [
        (
            True,
            'B --tree "((Santali_NM:1,Mundari_NM:1.1),(Hindi_IA:2,Sadri_IA:1.9)):3" '
            '--language-filters \'{"Name":".+"}\' --tree-label-property ID',
            lambda svg: True),
        (
            True,
            'B --tree "((Santali_NM:1,Mundari_NM:1.1),(Hindi_IA:2,Sadri_IA:1.9)):3" '
            '--tree-label-property ID',
            lambda svg: True),
        (
            True,
            'B --tree-dataset DATASET --tree-id 1 --tree-label-property ID',
            lambda svg: True),
        (
            True,
            'B --tree-dataset DATASET --tree-id 1 --glottocodes-as-tree-labels',
            lambda svg: True),
        (
            False,
            'param1 --tree abcd1234 --glottolog GLOTTOLOG',
            lambda svg: True),
        (
            False,
            'param1 --tree "((abcd1234:2,abcd1235:2.5)book1243:3):2;" --glottolog GLOTTOLOG',
            lambda svg: True),
        (
            False,
            'param1 --tree "((abcd1234:2,abcd1235:2.5):2,book1243:3):2;" --glottolog GLOTTOLOG',
            lambda svg: True),
    ]
)
def test_treemap(
        tmp_path, glottolog_dir, MetadataFreeStructureDataset, StructureDataset,
        with_full_dataset, opts, expect):
    out = tmp_path / 'test.svg'
    opts = opts.replace('DATASET', str(StructureDataset.directory))
    opts = opts.replace('GLOTTOLOG', str(glottolog_dir))
    print(opts)
    with warnings.catch_warnings():
        warnings.filterwarnings(
            'ignore', category=UserWarning, module='geopandas.plotting')
        runcli(
            'cldfviz.treemap',
            '{} --output {} --test {}'.format(
                StructureDataset.directory if with_full_dataset else MetadataFreeStructureDataset,
                out,
                opts))
    assert expect(out.read_text(encoding='utf8'))


def test_examples(ds_arg, tmp_path, capsys):
    main(['cldfviz.examples', ds_arg, '-o', str(tmp_path / 'ex.html')])
    assert tmp_path.joinpath('ex.html').exists()

    main(['cldfviz.examples', ds_arg])
    out, _ = capsys.readouterr()
    assert '<ol class="example">' in out


def test_erd(ds_arg, tmp_path, mocker):
    with pytest.raises(SystemExit):
        main(['cldfviz.erd', '-h'])

    with requests_mock.Mocker() as m:
        class Subprocess:
            @staticmethod
            def check_output(cmd, *args, **kw):
                out = None
                for i, token in enumerate(cmd):
                    if token == '-o':
                        out = cmd[i + 1]
                        break
                o = pathlib.Path(out) / 'diagrams' / 'summary' / 'relationships.real.large.svg'
                o.parent.mkdir(parents=True)
                o.write_text('a', encoding='utf8')
                return ''.encode('utf8')

        mocker.patch('cldfviz.commands.erd.subprocess', Subprocess())
        jar = tmp_path / 'jar'
        jar.write_text('abc', encoding='utf8')
        o = tmp_path / 'res.svg'
        m.get(requests_mock.ANY, text='abc')
        main(['cldfviz.erd', ds_arg, '--output', str(o), '--test', '--sqlite-jar', str(jar)])
        assert o.exists()


def test_network(capsys, StructureDataset, tmp_path):
    eattrs = tmp_path / 'e.py'
    eattrs.write_text(
        """dict(
    drop=lambda e: e.id == "2",
    color=lambda e: "red" if e.id == "1" else "blue",
    integer=lambda e: 1,
    float=lambda e: 2.1,
    boolean=True,
)""",
        encoding='utf8')

    nattrs = tmp_path / 'n.py'
    nattrs.write_text('dict(color=lambda n: "black" if n.id == "B" else None)', encoding='utf8')

    runcli(
        'cldfviz.network',
        '{} --edge-attributes {} --node-attributes {} --format graphml'.format(
            StructureDataset.directory,
            eattrs,
            nattrs,
        ))
    assert '<node ' in capsys.readouterr()[0]

    runcli('cldfviz.network', '{} --edge-attributes {}'.format(
        StructureDataset.directory,
        eattrs,
    ))
    assert 'digraph' in capsys.readouterr()[0]

    runcli('cldfviz.network', """{} --edge-filters '{}'""".format(
        StructureDataset.directory,
        '{"Description": "a"}'
    ))
    assert capsys.readouterr()[0].count('->') == 3

    runcli('cldfviz.network', """{} --edge-filters '{}'""".format(
        StructureDataset.directory,
        '{"Description": ["a"]}'
    ))
    assert capsys.readouterr()[0].count('->') == 3

    runcli('cldfviz.network', """{} --edge-filters '{}'""".format(
        StructureDataset.directory,
        '{"Description": "a|b"}'
    ))
    assert capsys.readouterr()[0].count('->') == 4

    runcli('cldfviz.network', """{} --edge-filters '{}'""".format(
        StructureDataset.directory,
        '{"iattr": 4}'
    ))
    assert capsys.readouterr()[0].count('->') == 1

    runcli('cldfviz.network', """{} --parameter 'H'""".format(StructureDataset.directory))
    assert capsys.readouterr()[0].count('->') == 2


@pytest.mark.parametrize(
    'ds,opts,expect',
    [
        (
            'StructureDataset',
            '--text-string "[](Source?with_the_works=false#cldf:__all__)"',
            lambda out: 'Peterson, John' in out),
        (
            'StructureDataset',
            '-l',
            lambda out: 'CodeTable' in out),
        (
            'StructureDataset',
            '--text-string "[ex](ExampleTable#cldf:igt)" --no-escape',
            lambda out: '<in>' in out),
        (
            'Wordlist',
            '--media-id 2',
            lambda out: 'The Text' in out),
    ]
)
def test_text(capsys, Wordlist, StructureDataset, ds, opts, expect):
    runcli('cldfviz.text', '{} --test {}'.format(
        (Wordlist if ds == 'Wordlist' else StructureDataset).directory, opts))
    assert expect(capsys.readouterr()[0])


def test_text_invalid(StructureDataset):
    with pytest.raises(ValueError):
        main([
            'cldfviz.text', '--text-string', '"a"', str(StructureDataset.directory / 'sources.bib')])


def test_text_multi_ds(ds_arg, capsys):
    main([
        'cldfviz.text',
        '--text-string', '"[](ParameterTable#cldf-p1:B) [](ParameterTable#cldf-p2:D)"',
        'p1:' + ds_arg,
        'p2:' + ds_arg,
    ])
    out = capsys.readouterr()[0]
    assert 'Gender' in out and 'Oblique' in out


def test_text_with_images(ds_arg, capsys, tmp_path):
    tmpl = tmp_path / 'templ.md'
    tmpl.write_text('![](map.html?parameters=B&pacific-centered#cldfviz.map-pref)')
    lmain(
        ['cldfviz.text', '--text-file', str(tmpl), '--test', '--output', str(tmp_path / 'test.md'), 'pref:' + ds_arg])
    assert tmp_path.joinpath('map.html').exists()

    main(['cldfviz.text', ds_arg, '--text-string', '![](tree.svg#cldfviz.tree)', '--test', '--output', str(tmp_path / 'test.md')])
    assert tmp_path.joinpath('tree.svg').exists()


class MF(MarkerFactory):
    def __init__(self, ds, args, param):
        MarkerFactory.__init__(self, ds, args, param)
        assert param == 'test'

    def __call__(self, map, language, values, colormaps):
        if self.args.format == 'html':
            return leaflet.LeafletMarkerSpec(tooltip='abcdefg', css='.leaflet-tooltip {}')
        return mpl.MPLMarkerSpec(text='abcdefg')


def test_map_misc(tmp_path, capsys, ds_arg, md_path_factory, MetadataFreeStructureDataset):
    def run(ds, opts):
        return functools.partial(runcli, 'cldfviz.map')(str(ds) + ' ' + opts)

    with pytest.raises(SystemExit):  # Shapes in colormap, but just one parameter.
        run(MetadataFreeStructureDataset,
            '--parameters param1 '
            '--colormaps \'{"val1": "circle", "val2": "#0a0", "val3": "#00a", "val4": "#000"}\'')

    with pytest.raises(SystemExit):  # Explicit color map for non-categorical parameter.
        run(ds_arg, '--parameters B --colormaps \'{"v":"#aaaaaa"}\'')

    with pytest.raises(SystemExit):  # Shapes in color maps for both parameters.
        run(ds_arg,
            '--parameters C,D '
            '--colormaps \'{"0":"circle","1":"diamond","2":"square"},'
            '{"0":"circle","1":"diamond","2":"square"}\'')

    with pytest.raises(SystemExit):  # Shapes in color maps, but more than 2 parameters.
        run(ds_arg,
            '--parameters C,D,E '
            '--colormaps \'{"0":"circle","1":"diamond","2":"square"},tol,tol\'')

    with pytest.raises(SystemExit):
        # Non-matching colormap values:
        run(MetadataFreeStructureDataset, '--parameters param1 --colormaps \'{"x": "y"}\'')
    out, _ = capsys.readouterr()
    assert 'ERROR' in out

    if WITH_CARTOPY:
        with warnings.catch_warnings():
            warnings.filterwarnings(
                'ignore', category=DeprecationWarning, module='cartopy.crs')
            # For JPG, there's a slightly different code path.
            run(ds_arg, '--format jpg --parameters C --output {}'.format(tmp_path / 'test.jpg'))


@pytest.mark.skipif(not WITH_CARTOPY, reason="Cannot run without cartopy.")
@pytest.mark.parametrize(
    'with_full_dataset,opts,expect_html,expect_svg',
    [
        (False, '--parameters param1', None, None),
        (
            False,
            '--parameters param1 '
            '--colormaps=\'{"val1":"#a00","val2":"#0a0","val3":"#00a","val4":"#000"}\'',
            None, None),
        (
            True,
            '--parameters C --overlay-geojson ecoregions',
            lambda html: 'ECO_NAME' in html, None),
        (
            True,
            '--parameters C --colormaps \'{"0":"circle","1":"diamond","2":"square"}\'',
            None, None),
        (
            True,
            '--parameters C,D --colormaps \'{"0":"circle","1":"diamond","2":"square"},tol\'',
            None, None),
        (
            'StructureDataset_listvalued_glottocode',
            '--parameters B,C --colormaps viridis,tol --language-properties Family_name '
            '--pacific-centered',
            None, None),
        (
            True,
            '--parameters C --language-filters \'{"Name":"Kharia"}\'',
            lambda html: 'Kharia' in html and 'Telugu' not in html, None),
        (
            True,
            '--parameters C --language-filters \'{"Filtered":1}\'',
            lambda html: 'Kharia' in html and 'Maithili' not in html, None),
        (
            True,
            '--parameters C --language-filters \'{"ListFiltered":"a"}\'',
            lambda html: 'Kharia' in html and 'Ho' not in html, None),
        (
            True,
            '--marker-factory cldfviz.map',
            None, None),
        (
            True,
            '--marker-factory {},test'.format(__file__),
            None, None),
        (
            False,
            '--parameters param1',
            None, lambda svg: 'param1' in svg),
        (
            True,
            '--title "The Title" --language-labels --parameters B,C --colormaps viridis,tol '
            '--language-properties Family_name --pacific-centered',
            None, lambda svg: 'The Title' in svg,
        ),
        (
            False,
            '--projection Robinson --parameters param1 --with-stock-img',
            None, None),
        (
            False,
            '--projection Robinson --parameters param1 --extent "50,150,50,-50"',
            None, None),
        (
            False,
            '--projection Robinson --parameters param1 --zorder \'{"v":5}\'',
            None, None),
        (
            False,
            '--parameters param1,param2 --no-borders',
            None, None),
        (
            True,
            '--parameters C,D --colormaps \'{"0":"circle","1":"diamond","2":"square"},tol\' '
            '--projection Mollweide',
            None, None),
        (
            True,
            '--parameters Z,C --projection Mollweide --pacific-centered',
            None, None),
    ]
)
def test_map(
        MetadataFreeStructureDataset, StructureDataset, md_path_factory,
        glottolog_dir, tmp_path, with_full_dataset, opts, expect_html, expect_svg):
    if isinstance(with_full_dataset, str):
        data = md_path_factory(with_full_dataset)
    else:
        data = StructureDataset.directory / 'StructureDataset-metadata.json' \
            if with_full_dataset else MetadataFreeStructureDataset

    def run(fmt):
        out = tmp_path / 'testmap.{}'.format(fmt)
        args = '{} --test --format {} --output {} --glottolog {} '.format(
            data, fmt, out, glottolog_dir)
        runcli('cldfviz.map', args + opts)
        assert out.exists()
        return out.read_text(encoding='utf8')

    html = run('html')
    if expect_html:
        assert expect_html(html)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            'ignore', category=DeprecationWarning, module='cartopy.crs')
        svg = run('svg')
    if expect_svg:
        assert expect_svg(svg)



import typing
import decimal
import functools
import itertools
import collections

import attr
import pycldf
from pycldf import orm
from pyglottolog.languoids import Languoid

from cldfviz.glottolog import Glottolog

CONTINUOUS = 1
CATEGORICAL = 2


class Language:
    def __init__(self, obj, glottolog: typing.Optional[Glottolog] = None):
        glottolog = glottolog or {}
        if isinstance(obj, str):
            obj = glottolog[obj]
            self.id = obj.id
            self.name = obj.name
            self.lat = obj.lat
            self.lon = obj.lon
        elif isinstance(obj, orm.Language):
            self.id = obj.id
            self.name = obj.name
            self.lat = obj.cldf.latitude
            self.lon = obj.cldf.longitude
            if self.lat is None and obj.cldf.glottocode in glottolog:
                # FIXME: If a language is mapped to multiple glottocodes, we could try to take the
                # midpoint of these as coordinate. (If longitudes have different signs, transform
                # back and forth appropriately, i.e. lon < 0 => lon = 360 - abs(lon))
                # shapely.geometry.MultiPoint([(0, 0), (1, 1)]).convex_hull.centroid
                self.lat = glottolog[obj.cldf.glottocode].lat
                self.lon = glottolog[obj.cldf.glottocode].lon
        else:
            raise TypeError(obj)
        self.lat = float(self.lat) if self.lat is not None else self.lat
        self.lon = float(self.lon) if self.lon is not None else self.lon


@attr.s
class Parameter:
    id = attr.ib()
    name = attr.ib()
    type = attr.ib(default=CATEGORICAL)
    domain = attr.ib(default=attr.Factory(dict))
    value_to_code = attr.ib(default=attr.Factory(dict))

    @classmethod
    def from_object(cls, obj):
        return cls(id=obj.id, name=getattr(obj.cldf, 'name', obj.id))


@functools.total_ordering
@attr.s(order=False, eq=False)
class Value:
    v = attr.ib()
    pid = attr.ib()
    lid = attr.ib()
    code = attr.ib()
    float = attr.ib(default=None)
    weight = attr.ib(default=None)

    def __attrs_post_init__(self):
        try:
            self.float = float(self.v)
        except (ValueError, TypeError):
            self.float = None

    def __eq__(self, other):
        return (self.lid, self.pid, self.v) == (other.lid, other.pid, other.v)

    def __lt__(self, other):
        return (self.lid, self.pid, self.v) < (other.lid, other.pid, other.v)

    @classmethod
    def from_row(cls, row, codes, weight_col=None):
        return cls(
            v=row.get('codeReference') or row['value'],
            lid=row['languageReference'],
            pid=row['parameterReference'],
            code=codes[row['parameterReference']][row['codeReference']]
            if row['parameterReference'] in codes and row['codeReference'] else None,
            weight=row[weight_col] if weight_col else None,
        )


class MultiParameter:
    """
    Extracts relevant data about a set of parameters from a CLDF dataset.

    :ivar parameters: `OrderedDict` mapping parameter IDs to :class:`Parameter` instances.
    """
    def __init__(self,
                 ds: pycldf.Dataset,
                 pids: typing.Iterable[str],
                 datatypes: typing.Iterable[str] = None,
                 include_missing: bool = False,
                 glottolog: typing.Optional[typing.Dict[str, typing.Union[dict, Languoid]]] = None,
                 language_properties: typing.Optional[typing.Iterable[str]] = None,
                 language_filter: typing.Optional[typing.Callable[[orm.Object], bool]] = None,
                 weight_col=None,
                 exclude_lang=None):
        self.include_missing = include_missing
        language_properties = language_properties or []

        if 'LanguageTable' in ds:
            langs = {lg.id: Language(lg, glottolog=glottolog)
                     for lg in ds.objects('LanguageTable')
                     if language_filter is None or language_filter(lg)}
        else:
            langs = {gc: Language(gc, glottolog=glottolog)
                     for gc in set(r['languageReference'] for r in
                                   ds.iter_rows('ValueTable', 'languageReference'))
                     if glottolog and gc in glottolog}

        langs = {
            k: v for k, v in langs.items() if v and (exclude_lang is None or not exclude_lang(v))}
        params = {p.id: Parameter.from_object(p)
                  for p in ds.objects('ParameterTable')} if 'ParameterTable' in ds else {}
        # For each pid, we add a parameter:
        self.parameters = collections.OrderedDict(
            [(pid, params.get(pid, Parameter(id=pid, name=pid))) for pid in pids])
        # For each language-property we add a parameter:
        for language_property in language_properties:
            self.parameters[language_property] = Parameter(
                id=language_property, name=language_property)
        if not self.parameters:
            # No parameters and no language property specified: Just plot language locations.
            self.parameters['__language__'] = Parameter(id='__language__', name='language')

        codes = collections.defaultdict(collections.OrderedDict)
        if 'CodeTable' in ds:
            for row in ds.iter_rows('CodeTable', 'id', 'parameterReference', 'name'):
                if row['parameterReference'] in self.parameters:
                    codes[row['parameterReference']][row['id']] = row['name']
        language_rows = []
        for i, language_property in enumerate(language_properties):
            if i == 0:
                language_rows = [
                    r for r in ds.iter_rows('LanguageTable', 'id', 'name')
                    if r['id'] in langs]
            if not all(isinstance(v[language_property], (int, float, decimal.Decimal))
                       for v in language_rows if v[language_property] is not None):
                counts = collections.Counter([r[language_property] for r in language_rows])
                codes[language_property] = collections.OrderedDict([
                    (p, p) for p in sorted(
                        set(r[language_property] for r in language_rows
                            if r[language_property] is not None),
                        key=lambda x: -counts[x])
                ])
        self.languages = collections.OrderedDict()
        self.values = []
        colmap = ['languageReference', 'parameterReference', 'value']
        if codes:
            colmap.append('codeReference')
        if pids:
            seen = {pid: False for pid in pids}
            comp = 'ValueTable' if ds.module == 'StructureDataset' else 'FormTable'
            for val in ds.iter_rows(comp, *colmap):
                seen[val['parameterReference']] = True
                if ((val['value'] is not None) or self.include_missing) and \
                        val['parameterReference'] in self.parameters:
                    lang = langs.get(val['languageReference'])
                    if lang:
                        self.languages[val['languageReference']] = lang
                        self.values.append(Value.from_row(val, codes, weight_col=weight_col))
                        self.parameters[val['parameterReference']] \
                            .value_to_code[str(val['value'])] = \
                            val.get('codeReference') or val['Value']
            if not all(seen[pid] for pid in pids):
                raise ValueError('Invalid parameter ID')
        for language_property in language_properties:
            for lang in language_rows:
                if (lang['id'] in langs) and lang[language_property] is not None:
                    if lang['id'] in langs:
                        self.languages.setdefault(lang['id'], langs[lang['id']])
                        self.values.append(Value(
                            v=lang[language_property],
                            pid=language_property,
                            lid=lang['id'],
                            code=language_property))
        if not self.values:
            # No parameters and no language property specified: Just plot language locations.
            if 'LanguageTable' not in ds:
                for lid, lang in langs.items():
                    self.languages.setdefault(lid, lang)
                    self.values.append(Value(
                        v='y',
                        pid='__language__',
                        lid=lid,
                        code='language'))
            else:
                for lang in ds.iter_rows('LanguageTable', 'id', 'name'):
                    if lang['id'] in langs:
                        self.languages.setdefault(lang['id'], langs[lang['id']])
                        self.values.append(Value(
                            v='y',
                            pid='__language__',
                            lid=lang['id'],
                            code='language'))

        for i, p in enumerate(self.parameters.values()):
            if p.id in codes:
                p.domain = codes[p.id]
            else:
                vals = [v for v in self.values if v.pid == p.id]
                if all(v.float is not None for v in vals) and \
                        (len(set(v.v for v in vals)) > 8 or  # noqa: W504
                         (datatypes and datatypes[i] == 'number')):
                    p.type = CONTINUOUS
                    p.domain = (min(v.float for v in vals), max(v.float for v in vals))
                else:
                    counts = collections.Counter([vv.v for vv in vals])
                    p.domain = collections.OrderedDict([
                        (v, v)
                        for v in sorted(set(vv.v for vv in vals), key=lambda vv: -counts[vv])])

    def __str__(self):  # pragma: no cover
        return str(self.parameters)

    def iter_languages(self) \
            -> typing.Iterator[typing.Tuple[Language, typing.Dict[str, typing.List[Value]]]]:
        for lid, values in itertools.groupby(sorted(self.values), lambda v: v.lid):
            values = {pid: list(vals) for pid, vals in itertools.groupby(values, lambda v: v.pid)}
            values = collections.OrderedDict(
                [(pid, values.get(pid, [])) for pid in self.parameters])
            if self.include_missing or all(bool(v) for v in values.values()):
                yield self.languages[lid], values



import sys
import copy
import typing
import pathlib
import textwrap
import xml.etree.cElementTree as ElementTree

import toytree
import toyplot.svg
from pycldf.trees import Tree
from newick import RESERVED_PUNCTUATION, Node
from clldutils.svg import pie, icon

from cldfviz.colormap import get_shape_and_color, SVG_SHAPE_MAP

__all__ = ['render']


def clean_node_label(s):
    if s:  # Automatically generated label mappings may map to `None`.
        for c in RESERVED_PUNCTUATION:
            s = s.replace(c, '_')
        return s.replace(' ', '_')
    return s


class SVGTree:
    def __init__(self, svg):
        self.svg = svg
        self.parent_map = {c: p for p in svg.iter() for c in p}

    @property
    def height(self):
        return float(self.svg.attrib['height'].replace('px', ''))

    @height.setter
    def height(self, val):
        self.svg.attrib['height'] = str(val) + 'px'
        viewBox = self.svg.attrib['viewBox'].split()
        viewBox[3] = str(val)
        self.svg.attrib['viewBox'] = ' '.join(viewBox)

    @property
    def width(self):
        return float(self.svg.attrib['width'].replace('px', ''))

    @width.setter
    def width(self, val):
        self.svg.attrib['width'] = str(val) + 'px'
        viewBox = self.svg.attrib['viewBox'].split()
        viewBox[2] = str(val)
        self.svg.attrib['viewBox'] = ' '.join(viewBox)

    @staticmethod
    def element(tag, parent, text=None, **attrs):
        ee = ElementTree.SubElement(parent, tag)
        ee.attrib = {k.rstrip('_').replace('_', '-'): str(v) for k, v in attrs.items()}
        if text:
            ee.text = text
        return ee

    @staticmethod
    def marker(parent, weighted_colors):
        res = get_shape_and_color(weighted_colors)
        if res:
            g = ElementTree.SubElement(parent, 'g')
            g.attrib['transform'] = 'scale(0.5)'
            parent = g
            p = ElementTree.fromstring(icon(res[1].replace('#', SVG_SHAPE_MAP[res[0]])))
        else:
            ratios, colors = [c[0] for c in weighted_colors], [c[1] for c in weighted_colors]
            p = ElementTree.fromstring(pie(ratios, colors, width=20, stroke_circle=True))
        parent.extend(p.findall('./{}path'.format('{http://www.w3.org/2000/svg}')))
        parent.extend(p.findall('./{}circle'.format('{http://www.w3.org/2000/svg}')))

    def visit_leafs(self, visitor, *args, **kw):
        for t in self.svg.findall('.//g[@class="toytree-TipLabels"]/g/text'):
            visitor(self, t, self.parent_map[t], *args, **kw)

    def __bytes__(self):
        kw = dict(encoding='utf8')
        if sys.version_info >= (3, 8):
            kw['xml_declaration'] = True
        return ElementTree.tostring(self.svg, **kw)

    def __str__(self):
        return bytes(self).decode('utf8')


def render(nwk: typing.Union[Node, Tree],
           tree_object: typing.Optional[Tree] = None,
           output: typing.Optional[pathlib.Path] = None,
           glottolog_mapping: typing.Optional[typing.Dict[str, typing.Tuple[str, str]]] = None,
           legend: typing.Optional[str] = None,
           width: typing.Optional[int] = 500,
           height: typing.Optional[int] = None,
           styles: typing.Optional[dict] = None,
           with_glottolog_links: bool = False,
           labels: typing.Optional[typing.Union[typing.Callable[[Node], str], dict]] = None,
           leafs: typing.Optional[typing.Union[typing.Callable[[Node], bool], list]] = None,
           data=None,
           ) -> typing.Union[pathlib.Path, str]:
    glottolog_mapping = glottolog_mapping or {}
    if isinstance(nwk, Tree):
        tree_object = nwk
        nwk = tree_object.newick(strip_comments=True)

    def rename(n):
        if n.name in glottolog_mapping:
            n.name = "{}--{}".format(n.name, glottolog_mapping[n.name][0])
        if not n.is_leaf:
            n.name = None

    def rename2(n):
        n.name = clean_node_label(labels(n) if callable(labels) else labels[n.name]) \
            if n.name else n.name

    if leafs:
        if callable(leafs):
            leafs = [n.name for n in nwk.walk() if n.name and leafs(n)]
        nwk.prune_by_names(leafs, inverse=True)
    if data:
        nwk.prune_by_names(list(data.values.keys()), inverse=True)
    if with_glottolog_links:
        nwk.visit(rename)
    if labels and (not data):
        nwk.visit(rename2)

    def pad(n):
        if n.name and n.is_leaf:
            n.name = n.name + '#############'  # FIXME: pad to fit longest label
    nwk.visit(pad)

    style = dict(
        width=width,
        height=height or sum(1 for n in nwk.walk() if n.is_leaf) * (23 if data else 15) + 150,
        node_hover=True,
        tip_labels_align=True,
        tip_labels_style={
            "fill": "#262626",
            "font-size": "11px",
            "-toyplot-anchor-shift": "15px",
            "line-height": "14px",
        },
        scalebar=bool(getattr(tree_object, 'tree_branch_length_unit', None)) or bool(legend),
    )
    style.update(styles or {})
    canvas, axes, mark = toytree.tree(nwk.newick + ";", tree_format=1).draw(**style)
    if legend:
        axes.label.text = legend
    res = SVGTree(toyplot.svg.render(canvas, None))
    if with_glottolog_links:
        res.visit_leafs(add_glottolog_links, glottolog_mapping)

    if data:
        res.visit_leafs(add_marker, data, labels)
        add_legend(res, data)
    else:
        res.visit_leafs(
            lambda s, t, p: setattr(t, 'text', t.text.rstrip('#') if t.text else t.text))

    if output:
        output.write_bytes(bytes(res))
        return output
    return str(res)


def add_glottolog_links(svg, t, _, gcodes):
    "Post-process the SVG to turn leaf names with Glottocodes into links"""
    if t.text:
        lid, _, gcode = t.text.strip().partition('--')
        if gcode:
            se = svg.element('text', t, **copy.copy(t.attrib))
            gname = gcodes[lid][1]
            if gname:
                se.text = '{} - {} [{}]'.format(lid, gname, gcode)
            else:
                se.text = '{} - [{}]'.format(lid, gcode)
            se.attrib['fill'] = '#0000ff'
            t.tag = 'a'
            t.attrib = {
                'href': 'https://glottolog.org/resource/languoid/id/{}'.format(gcode),
                # 'title': 'The glottolog name',
            }
            t.text = None


def add_marker(svg, t, parent, data, labels):
    t.text = t.text.rstrip('#') if t.text else t.text
    if t.text in data.values:
        t.attrib['x'] = str(float(t.attrib['x']) + 15)

        g = ElementTree.SubElement(parent, 'g')
        g.attrib = dict(transform="translate(0,-10)")
        svg.marker(g, data.values[t.text])
        if t.text in (labels or {}):
            t.text = labels[t.text]


def add_legend(svg, data):
    def shorten(text, width):
        return textwrap.shorten(str(text), width, placeholder='…')

    def row(legend, y, weighted_colors, label, **attrs):
        row_ = svg.element('g', legend, transform="translate(10,{})".format(y))
        if weighted_colors:
            svg.marker(row_, weighted_colors)
        svg.element(
            'text', row_,
            x=30 if weighted_colors else 0, y=15,
            text=shorten(label, 25 if weighted_colors else 30), stroke_width=0, **attrs)

    y = 0
    legend = svg.element(
        'g', svg.svg,
        transform="translate({},{})".format(svg.width - 20, 45), style="font-size: 12px")
    rect = svg.element('rect', legend, x=0, y=0, width='200', height=svg.height, rx=5, fill='white')
    pid_with_color = None
    if any(cm.with_shapes for cm in data.colormaps.values()):
        for pid, cm in data.colormaps.items():
            if not cm.with_shapes:
                pid_with_color = pid
                break

    for i, (pid, parameter) in enumerate(data.parameters.items()):
        if i != 0:
            svg.element('line', legend, x1=5, y1=y, x2=195, y2=y, stroke='black')
            y += 3
        else:
            y += 5

        row(legend,
            y,
            None,
            parameter.name,
            font_weight='bold')
        y += 25
        if isinstance(parameter.domain, tuple):
            min_, max_ = parameter.domain
            row_ = svg.element('g', legend, transform="translate(10,{})".format(y))
            svg.element('text', row_, x=0, y=15, text=str(min_), stroke_width=0)
            svg.element(
                'text', row_, x=180, y=15, text=str(max_), text_anchor='end', stroke_width=0)
            y += 25
            row_ = svg.element('g', legend, transform="translate(10,{})".format(y))
            for i in range(10):
                svg.element(
                    'rect',
                    row_,
                    x=i * 18,
                    y=0,
                    width='18', height='18',
                    fill=data.colormaps[pid](min_ + i * (max_ - min_) / 10))
            y += 25
        else:
            for v, label in parameter.domain.items():
                weighted_colors = [
                    (1, data.colormaps[pid](v) if j == i else '#ffffff')
                    for j in range(len(data.parameters))]
                if pid_with_color == pid:
                    weighted_colors = [(1, data.colormaps[pid](v))]
                row(legend, y, weighted_colors, label)
                y += 25
    rect.attrib['height'] = str(y)

    svg.width = svg.width + 220
    svg.height = max([svg.height, 45 + y])



"""
Map plotting with matplotlib and cartopy
"""
import json
import textwrap
import warnings

import attr
import numpy as np
import cartopy.feature
import cartopy.crs
from matplotlib import pyplot as plt
from matplotlib.patches import Wedge, Rectangle, Circle
from matplotlib.legend_handler import HandlerPatch
from PIL import Image

from cldfviz.colormap import get_shape_and_color, weighted_colors
from .base import Map, PACIFIC_CENTERED

SHAPE_MAP = {
    'triangle_down': '^',
    'triangle_up': 'v',
    'square': 's',
    'diamond': 'D',
    'circle': 'o',
}


def iter_subclasses(cls):
    for cls_ in cls.__subclasses__():
        yield cls_
        yield from iter_subclasses(cls_)


class HandleWedge(HandlerPatch):
    def create_artists(
            self,
            legend,
            orig_handle,
            xdescent,
            ydescent,
            width,
            height,
            fontsize,
            trans):
        center = 0.5 * width - 0.5 * xdescent, 0.5 * height - 0.5 * ydescent
        p = Wedge(
            center, height / 1.5, orig_handle.theta1, orig_handle.theta2, width=orig_handle.width)
        self.update_prop(p, orig_handle, legend)
        p.set_transform(trans)
        return [p]


@attr.s
class MPLMarkerSpec:
    marker_kw = attr.ib(default=attr.Factory(dict))
    text = attr.ib(default=None)
    text_offset_x = attr.ib(default=None)
    text_offset_y = attr.ib(default=None)
    text_kw = attr.ib(default=attr.Factory(dict))


class MapPlot(Map):
    """
    A geographic map, zoomed in to the extent defined by a list of languages.
    """
    __formats__ = ['jpg', 'png', 'pdf', 'svg']
    __marker_class__ = MPLMarkerSpec

    def __init__(self, languages, args):
        Map.__init__(self, languages, args)
        lats = [k.lat for k in languages if k.lat is not None]
        lons = [k.lon for k in languages if k.lon is not None]
        self.central_longitude = PACIFIC_CENTERED if args.pacific_centered else 0

        if args.extent:
            left, right, top, bottom = [float(x) for x in args.extent.replace('"', '').split(',')]
            self.extent = [left, right, bottom, top]
        else:
            self.extent = [
                round(min(lons) - args.padding_left, 1) + self.central_longitude,
                round(max(lons) + args.padding_right, 1) + self.central_longitude,
                round(min(lats) - args.padding_bottom, 1),
                round(max(lats) + args.padding_top, 1)
            ]
        self.ax = None
        self.scaling_factor = 1
        self.proj = getattr(cartopy.crs, args.projection)(central_longitude=self.central_longitude)

    def __enter__(self):
        plt.clf()
        fig = plt.figure(figsize=(self.args.width, self.args.height), dpi=self.args.dpi)
        ax = fig.add_subplot(1, 1, 1, projection=self.proj)
        if self.args.projection == 'PlateCarree':
            ax.set_extent(self.extent, crs=self.proj)
        else:
            pass
        if self.args.with_stock_img:
            ax.stock_img()
        if (not self.args.test) and (not self.args.with_stock_img):  # pragma: no cover
            ax.coastlines(resolution='50m', color='darkgrey')
            ax.add_feature(cartopy.feature.LAND, color='beige', zorder=1)
            if self.args.with_ocean:
                ax.add_feature(cartopy.feature.OCEAN, color='#97B5E1', zorder=2)
            if not self.args.no_borders:
                ax.add_feature(cartopy.feature.BORDERS, linestyle=':', zorder=4)
            ax.add_feature(cartopy.feature.LAKES, color="#97B5E1", alpha=0.5, zorder=3)
            ax.add_feature(cartopy.feature.RIVERS, color="#97B5E1", zorder=3)
        self.ax = ax
        # Figure out the scaling factor between degrees and pixels. Many config values are given
        # in pixels but need to be converted to degrees for plotting.
        m = self.proj._as_mpl_transform(axes=self.ax)
        y0 = m.transform_point((self.extent[0], self.extent[2])).tolist()[1]
        y1 = m.transform_point((self.extent[0], self.extent[2] + 1)).tolist()[1]
        self.scaling_factor = 1.0 / (y1 - y0)
        # So 1 px = self.scaling_factor * 1°
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.args.title:
            plt.title(self.args.title)
        format = self.args.output.suffix.replace('.', '').lower()
        if format == 'jpg':
            mplfname = self.args.output.parent / '{}.png'.format(self.args.output.stem)
            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', category=UserWarning, module='cartopy.mpl.style')
                plt.savefig(str(mplfname), bbox_inches="tight")
            img = Image.open(str(mplfname)).convert('RGB')
            img.save(str(self.args.output), optimize=True, quality=95)
            mplfname.unlink()
        else:
            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', category=UserWarning, module='cartopy.mpl.style')
                plt.savefig(str(self.args.output), bbox_inches="tight")

        plt.close()

    @staticmethod
    def add_options(parser, help_suffix):
        for direction in ['left', 'right', 'top', 'bottom']:
            parser.add_argument(
                '--padding-{}'.format(direction),
                help="{} padding of the map in degrees. {}".format(
                    direction.capitalize(), help_suffix),
                type=int,
                default=1,
            )
        parser.add_argument(
            '--extent',
            help="Set extent of the figure in terms of coordinates (left, right, top, bottom)",
            default=None
        )
        parser.add_argument(
            '--width',
            help="Width of the figure in inches. {}".format(help_suffix),
            type=float,
            default=6.4,
        )
        parser.add_argument(
            '--height',
            help="Height of the figure in inches. {}".format(help_suffix),
            type=float,
            default=4.8,
        )
        parser.add_argument(
            '--dpi',
            help="Pixel density of the figure. {}".format(help_suffix),
            type=float,
            default=100.0,
        )
        parser.add_argument(
            '--projection',
            help="Map projection. For details, see "
                 "https://scitools.org.uk/cartopy/docs/latest/crs/projections.html "
                 "Note that not all options work for all projections; e.g. --language-labels "
                 "will only work with PlateCarree, the default projection."
                 "{}".format(help_suffix),
            choices=[
                c.__name__ for c in iter_subclasses(cartopy.crs.Projection)
                if c.__module__ == 'cartopy.crs' and not c.__name__.startswith('_')],
            default='PlateCarree',
        )
        parser.add_argument(
            '--with-ocean',
            help="Render the oceans in the same color as other bodies of water. Since cartopy's "
                 "OCEAN polygon is somewhat broken, this may result in broken SVG output. Thus, we "
                 "don't include this by default. {}".format(help_suffix),
            action="store_true",
            default=False,
        )
        parser.add_argument(
            '--with-stock-img',
            help="Add a map underlay (using cartopy's `stock_img` method). {}".format(help_suffix),
            action="store_true",
            default=False,
        )
        parser.add_argument(
            '--no-borders',
            help="Do not add country borders to the map. {}".format(help_suffix),
            action="store_true",
            default=False,
        )
        parser.add_argument(
            '--zorder',
            help="Determine zorder of individual markers by color.",
            type=lambda s: json.loads(s),
            action="store",
            default={},
        )

    def _lonlat(self, language):
        lat, lon = language.lat, language.lon
        if self.central_longitude:
            lon = lon - self.central_longitude
            if lon < -180:
                lon += 360
        return lon, lat

    def pie_markers(self, colors):
        if len(colors) == 1:
            # Draw a full circle
            x = np.cos(np.linspace(0, 2 * np.pi, 30)).tolist()
            y = np.sin(np.linspace(0, 2 * np.pi, 30)).tolist()
            yield colors[0][1], np.column_stack([x, y])
            return
        start = 0.
        for ratio, color in colors:
            x = [0] + \
                np.cos(np.linspace(2 * np.pi * start, 2 * np.pi * (start + ratio), 30)).tolist() + \
                [0]
            y = [0] + \
                np.sin(np.linspace(2 * np.pi * start, 2 * np.pi * (start + ratio), 30)).tolist() + \
                [0]
            yield color, np.column_stack([x, y])
            start += ratio

    def add_language(self, language, values, colormaps, spec=None):
        # add zorder by using a point-system that penalizes missing data
        # according to user-defined weights
        if self.args.zorder:
            zorders = []
            for val in values.values():
                zorders += [self.args.zorder.get(val[0].v.split('-')[-1], 5)]
            zorder = sum(zorders)
        else:
            zorder = 5 * len(values)

        if spec:
            marker_kw = dict(
                color='white',
                markersize=self.args.markersize,
                zorder=20,
                marker='o',
                markeredgecolor='black',
                linewidth=1,
                transform=cartopy.crs.Geodetic(),
            )
            marker_kw.update(spec.marker_kw)
            self.ax.plot(language.lon, language.lat, **marker_kw)
            if spec.text:
                text_kw = dict(zorder=20, fontsize='small')
                text_kw.update(spec.text_kw)
                self.ax.text(
                    language.lon + (spec.text_offset_x or 0),
                    language.lat + (spec.text_offset_y or 0),
                    spec.text,
                    **text_kw)
            return
        lon, lat = self._lonlat(language)
        colors = weighted_colors(values, colormaps)

        if self.args.projection != 'PlateCarree':
            res = get_shape_and_color(colors)
            if res:
                self.ax.plot(
                    language.lon, language.lat,
                    color=res[1],
                    markersize=self.args.markersize,
                    zorder=zorder,
                    marker=SHAPE_MAP[res[0]],
                    markeredgecolor='black',
                    linewidth=1,
                    transform=cartopy.crs.Geodetic(),
                )
                return

            # Use scatter to create pie-markers suitable for the projection.
            for color, marker in self.pie_markers(colors):
                self.ax.scatter(
                    [language.lon], [language.lat],
                    marker=marker,
                    s=[self.args.markersize * 10],
                    transform=cartopy.crs.Geodetic(),
                    linewidth=1,
                    edgecolor='black',
                    zorder=zorder,
                    facecolor=color,
                )
            return

        res = get_shape_and_color(colors)
        if res:
            self.ax.plot(
                lon, lat,
                color=res[1],
                markersize=self.args.markersize,
                zorder=zorder,
                marker=SHAPE_MAP[res[0]],
                markeredgecolor='black',
                linewidth=1,
            )
        else:
            s = 0
            for ratio, color in colors:
                angle = 360.0 * ratio
                self.ax.add_patch(Wedge(
                    [lon, lat],
                    self.args.markersize * self.scaling_factor / 2.0,
                    s,
                    s + angle,
                    facecolor=color,
                    linewidth=0,
                    label=language.name,
                    zorder=zorder
                ))
                s += angle
            self.ax.add_patch(Circle(
                (lon, lat),
                self.args.markersize * self.scaling_factor / 2.0,
                fill=False,
                edgecolor="black",
                linewidth=1,
                zorder=zorder))
        if self.args.language_labels:
            self.ax.text(
                lon + self.args.markersize * self.scaling_factor + 3 * self.scaling_factor,
                lat, language.name,
                zorder=zorder + 10,
                fontsize='small')

    def add_legend(self, parameters, colormaps):
        def wrapped_label(s):
            return '\n'.join(textwrap.wrap(s, width=20))

        with_shapes = False
        if 1 <= len(parameters) <= 2:
            for pid, parameter in parameters.items():
                if not isinstance(parameter.domain, tuple):
                    for v, label in parameter.domain.items():
                        if colormaps[pid](v) in SHAPE_MAP:
                            with_shapes = True
                            break

        if with_shapes:
            handles = []
            for pid, parameter in parameters.items():
                handles.append(
                    Rectangle(
                        (0, 0), 1, 1, fc="w", fill=False, edgecolor='none', linewidth=0,
                        label=wrapped_label(parameter.name)))
                for v, label in parameter.domain.items():
                    color = colormaps[pid](v)
                    handles.append(
                        plt.Line2D(
                            [], [],
                            marker=SHAPE_MAP[color] if color in SHAPE_MAP else 'o',
                            color='#000000' if color in SHAPE_MAP else color,
                            linewidth=1,
                            linestyle='',
                            label=wrapped_label(label))
                    )
            self.ax.legend(
                bbox_to_anchor=(1, 1),
                handles=handles,
                loc='upper left',
            )
            return

        handles = []
        s, angle = 0, 360.0 / len(parameters)
        for pid, parameter in parameters.items():
            handles.append(Wedge(
                [-100, -100],
                self.args.markersize,
                s,
                s + angle,
                facecolor="None",
                edgecolor="black",
                label=wrapped_label(parameter.name),
                zorder=20
            ))
            if isinstance(parameter.domain, tuple):
                cbar = plt.colorbar(
                    colormaps[pid].scalar_mappable(),
                    ax=self.ax,
                    aspect=80,
                    label=parameter.name,
                    shrink=0.5,
                    pad=0.05,
                    location="bottom",
                )
                cbar.set_ticks([0.0, 1.0])
                cbar.set_ticklabels([
                    str(round(parameter.domain[0], 2)),
                    str(round(parameter.domain[1], 2))])
            else:
                for v, label in parameter.domain.items():
                    handles.append(Wedge(
                        [-100, -100],
                        self.args.markersize,
                        s,
                        s + angle,
                        facecolor=colormaps[pid](v),
                        edgecolor="black",
                        label=wrapped_label(label),
                        zorder=20
                    ))
            s += angle
        self.ax.legend(
            bbox_to_anchor=(1, 1),
            handles=handles,
            loc='upper left',
            handler_map={Wedge: HandleWedge()})
